@@ -28,6 +28,7 @@ import email.charset
 import fcntl
 import fnmatch
 import gettext
+import grp
 import locale
 import logging
 import re
@@ -73,6 +74,20 @@ PID_FILE = "/var/run/unattended-upgrades.pid"
 # set from the sigint signal handler
 SIGNAL_STOP_REQUEST = False
 
+class LoggingDateTime:
+    """The date/time representation for the dpkg log file timestamps"""
+    LOG_DATE_TIME_FMT = "%Y-%m-%d  %H:%M:%S"
+
+    @classmethod
+    def as_string(cls):
+        """Return the current date and time as LOG_DATE_TIME_FMT string"""
+        return datetime.datetime.now().strftime(cls.LOG_DATE_TIME_FMT)
+
+    @classmethod
+    def from_string(cls, logstr):
+        """Take a LOG_DATE_TIME_FMT string and return datetime object"""
+        return datetime.datetime.strptime(logstr, cls.LOG_DATE_TIME_FMT)
+
 
 class UnknownMatcherError(ValueError):
     pass
@@ -95,7 +110,9 @@ class UnattendedUpgradesCache(apt.Cache):
         """ Adjust candidate versions to match highest allowed origin
 
             This adjusts the origin even if the candidate has a higher
-            version
+            version. This is needed when e.g. a package is available in
+            the security pocket but there is also a package in the
+            updates pocket with a higher version number
         """
         for pkg in self:
             # important! this avoids downgrades below
@@ -109,9 +126,10 @@ class UnattendedUpgradesCache(apt.Cache):
             # not the candidate
             new_cand = None
             for ver in pkg.versions:
-                # ignore versions that the user marked with priority < 0
+                # ignore versions that the user marked with priority < 100
                 # (and ensure we have a python-apt that supports this)
-                if hasattr(ver, "policy_priority") and ver.policy_priority < 0:
+                if (hasattr(ver, "policy_priority") and
+                        ver.policy_priority < 100):
                     logging.debug("ignoring ver '%s' with priority < 0" % ver)
                     continue
                 if is_allowed_origin(ver, self.allowed_origins):
@@ -182,11 +200,20 @@ class LogInstallProgress(apt.progress.base.InstallProgress):
     def _redirect_output(self):
         # do not create log in dry-run mode, just output to stdout/stderr
         if not apt_pkg.config.find_b("Debug::pkgDPkgPM", False):
-            logfd = os.open(
-                self.logfile_dpkg, os.O_RDWR | os.O_APPEND | os.O_CREAT, 0o644)
+            logfd = self._get_logfile_dpkg_fd()
             os.dup2(logfd, 1)
             os.dup2(logfd, 2)
 
+    def _get_logfile_dpkg_fd(self):
+        logfd = os.open(
+            self.logfile_dpkg, os.O_RDWR | os.O_APPEND | os.O_CREAT, 0o640)
+        try:
+            adm_gid = grp.getgrnam("adm").gr_gid
+            os.fchown(logfd, 0, adm_gid)
+        except KeyError:
+            pass
+        return logfd
+
     def update_interface(self):
         # call super class first
         apt.progress.base.InstallProgress.update_interface(self)
@@ -200,6 +227,7 @@ class LogInstallProgress(apt.progress.base.InstallProgress):
         if self.verbose:
             if self.output_logfd is None:
                 self.output_logfd = os.open(self.logfile_dpkg, os.O_RDONLY)
+                os.lseek(self.output_logfd, 0, os.SEEK_END)
             try:
                 select.select([self.output_logfd], [], [], 0)
             except select.error as e:
@@ -208,7 +236,16 @@ class LogInstallProgress(apt.progress.base.InstallProgress):
             # output to stdout in verbose mode only
             os.write(1, os.read(self.output_logfd, 1024))
 
+    def _log_in_dpkg_log(self, msg):
+        logfd = self._get_logfile_dpkg_fd()
+        os.write(logfd, msg.encode("utf-8"))
+        os.close(logfd)
+
+    def finish_update(self):
+        self._log_in_dpkg_log("Log ended: %s\n\n" % LoggingDateTime.as_string())
+
     def fork(self):
+        self._log_in_dpkg_log("Log started: %s\n" % LoggingDateTime.as_string())
         pid = os.fork()
         if pid == 0:
             self._fixup_fds()
@@ -341,8 +378,8 @@ def match_whitelist_string(whitelist, origin):
             raise UnknownMatcherError(
                 "Unknown whitelist entry for macher '%s' (token '%s')" % (
                     what, token))
-        logging.debug("matching '%s'='%s' against '%s'" % (
-                      what, value, origin))
+        #logging.debug("matching '%s'='%s' against '%s'" % (
+        #              what, value, origin))
     return res
 
 
@@ -368,7 +405,7 @@ def upgrade_normal(cache, pkgs_to_upgrade, logfile_dpkg, verbose):
 
 
 def upgrade_in_minimal_steps(cache, pkgs_to_upgrade, blacklist,
-                             logfile_dpkg="", verbose=False):
+                             whitelist, logfile_dpkg="", verbose=False):
 
     install_log = LogInstallProgress(logfile_dpkg, verbose)
     install_log.progress_log += ".minimal-steps"
@@ -397,7 +434,8 @@ def upgrade_in_minimal_steps(cache, pkgs_to_upgrade, blacklist,
                 continue
             # double check that we are not running into side effects like
             # what could have been caused LP: #1020680
-            if not check_changes_for_sanity(cache, allowed_origins, blacklist):
+            if not check_changes_for_sanity(cache, allowed_origins, blacklist,
+                                            whitelist):
                 logging.info("While building minimal partition: "
                              "cache has not allowed changes")
                 cache.clear()
@@ -468,8 +506,19 @@ def is_pkgname_in_blacklist(pkgname, blacklist):
             return True
     return False
 
+def is_pkgname_in_whitelist(pkgname, whitelist):
+    # a empty whitelist means the user does not want to use this feature
+    if not whitelist:
+        return True
+    for whitelist_regexp in whitelist:
+        if re.match(whitelist_regexp, pkgname):
+            logging.debug("only upgrading the following package '%s'" %
+                          pkgname)
+            return True
+    return False
 
-def check_changes_for_sanity(cache, allowed_origins, blacklist):
+
+def check_changes_for_sanity(cache, allowed_origins, blacklist, whitelist):
     if cache._depcache.broken_count != 0:
         return False
     for pkg in cache:
@@ -481,6 +530,19 @@ def check_changes_for_sanity(cache, allowed_origins, blacklist):
                 logging.debug("pkg '%s' not in allowed origin" % pkg.name)
                 return False
             if is_pkgname_in_blacklist(pkg.name, blacklist):
+                logging.debug("pkg '%s' package has been blacklisted" %
+                              pkg.name)
+                return False
+            # a strict whitelist will not allow any changes not in the 
+            # whitelist, most people will want the relaxed whitelist
+            # that whitelists a package but pulls in the package 
+            # dependencies
+            strict_whitelist = apt_pkg.config.find_b(
+                "Unattended-Upgrade::Package-Whitelist-Strict", False)
+            if (strict_whitelist and
+                not is_pkgname_in_whitelist(pkg.name, whitelist)):
+                logging.debug("pkg '%s' package is not whitelisted" %
+                              pkg.name)
                 return False
             if pkg._pkg.selected_state == apt_pkg.SELSTATE_HOLD:
                 logging.debug("pkg '%s' is on hold" % pkg.name)
@@ -697,8 +759,8 @@ def _send_mail_using_sendmail(to_address, subject, body):
     return ret
 
 
-def send_summary_mail(pkgs, res, pkgs_kept_back, mem_log, logfile_dpkg):
-    " send mail (if configured in Unattended-Upgrade::Mail) "
+def send_summary_mail(pkgs, res, pkgs_kept_back, mem_log, dpkg_log_content):
+    """ send mail (if configured in Unattended-Upgrade::Mail) """
     to_email = apt_pkg.config.find("Unattended-Upgrade::Mail", "")
     if not to_email:
         return
@@ -713,7 +775,7 @@ def send_summary_mail(pkgs, res, pkgs_kept_back, mem_log, logfile_dpkg):
             "Unattended-Upgrade::MailOnlyOnError", False)):
         return
     # Check if reboot-required flag is present
-    logging.debug("Sending mail with '%s' to '%s'" % (logfile_dpkg, to_email))
+    logging.debug("Sending mail to '%s'" % to_email)
     if os.path.isfile(REBOOT_REQUIRED_FILE):
         subject = _(
             "[reboot required] unattended-upgrades result for '%s': %s") % (
@@ -736,10 +798,9 @@ def send_summary_mail(pkgs, res, pkgs_kept_back, mem_log, logfile_dpkg):
         body += " " + wrap(" ".join(pkgs_kept_back), 70, " ")
         body += "\n"
     body += "\n"
-    if os.path.exists(logfile_dpkg):
+    if dpkg_log_content:
         body += _("Package installation log:") + "\n"
-        with open(logfile_dpkg) as fp:
-            body += fp.read()
+        body += dpkg_log_content
         body += "\n\n"
     body += _("Unattended-upgrades log:\n")
     body += mem_log.getvalue()
@@ -755,14 +816,20 @@ def send_summary_mail(pkgs, res, pkgs_kept_back, mem_log, logfile_dpkg):
     logging.debug("mail returned: %s" % ret)
 
 
-def do_install(cache, pkgs_to_upgrade, blacklisted_pkgs, options,
-               logfile_dpkg):
+def do_install(cache, pkgs_to_upgrade, blacklisted_pkgs, whitelisted_pkgs,
+               options, logfile_dpkg):
     # set debconf to NON_INTERACTIVE, redirect output
     os.putenv("DEBIAN_FRONTEND", "noninteractive")
     setup_apt_listchanges()
 
     logging.info(_("Writing dpkg log to '%s'") % logfile_dpkg)
 
+    marked_delete = [pkg for pkg in cache.get_changes() if pkg.marked_delete]
+    if marked_delete:
+        raise AssertionError(
+            "Internal error. The following packages are marked for "
+            "removal:%s" % "".join([pkg.name for pkg in marked_delete]))
+
     pkg_install_success = False
     try:
         if (options.minimal_upgrade_steps or
@@ -778,7 +845,7 @@ def do_install(cache, pkgs_to_upgrade, blacklisted_pkgs, options,
             # try upgrade all "pkgs" in minimal steps
             pkg_install_success = upgrade_in_minimal_steps(
                 cache, [pkg.name for pkg in pkgs_to_upgrade],
-                blacklisted_pkgs, logfile_dpkg,
+                blacklisted_pkgs, whitelisted_pkgs, logfile_dpkg,
                 options.verbose or options.debug)
         else:
             pkg_install_success = upgrade_normal(
@@ -832,7 +899,6 @@ def _setup_logging(options):
             # COMPAT only
             apt_pkg.config.find("APT::UnattendedUpgrades::LogFile",
                                 "unattended-upgrades.log")))
-
     if not options.dry_run and not os.path.exists(logdir):
         os.makedirs(logdir)
 
@@ -863,6 +929,10 @@ def get_blacklisted_pkgs():
     return apt_pkg.config.value_list("Unattended-Upgrade::Package-Blacklist")
 
 
+def get_whitelisted_pkgs():
+    return apt_pkg.config.value_list("Unattended-Upgrade::Package-Whitelist")
+
+
 def reboot_if_requested_and_needed(shutdown_lock=0):
     # auto-reboot (if required and the config for this is set
     if (apt_pkg.config.find_b(
@@ -883,46 +953,80 @@ def write_stamp_file():
         pass
 
 
+def try_to_upgrade(pkg, pkgs_to_upgrade, pkgs_kept_back, cache, allowed_origins,
+                     blacklisted_pkgs, whitelisted_pkgs):
+
+    try:
+        pkg.mark_upgrade()
+        if check_changes_for_sanity(cache, allowed_origins,
+                                    blacklisted_pkgs, whitelisted_pkgs):
+            # add to packages to upgrade
+            pkgs_to_upgrade.append(pkg)
+            # re-eval pkgs_kept_back as the resolver may fail to
+            # directly upgrade a pkg, but that may work during
+            # a subsequent operation, see debian bug #639840
+            for pkgname in pkgs_kept_back:
+                if (cache[pkgname].marked_install or
+                        cache[pkgname].marked_upgrade):
+                    pkgs_kept_back.remove(pkgname)
+                    pkgs_to_upgrade.append(cache[pkgname])
+        else:
+            logging.debug("sanity check failed")
+            rewind_cache(cache, pkgs_to_upgrade)
+            pkgs_kept_back.append(pkg.name)
+    except SystemError as e:
+        # can't upgrade
+        logging.warning(
+            _("package '%s' upgradable but fails to "
+                "be marked for upgrade (%s)") % (pkg.name, e))
+        rewind_cache(cache, pkgs_to_upgrade)
+        pkgs_kept_back.append(pkg.name)
+
+
 def calculate_upgradable_pkgs(cache, options, allowed_origins,
-                              blacklisted_pkgs):
+                              blacklisted_pkgs, whitelisted_pkgs):
     pkgs_to_upgrade = []
     pkgs_kept_back = []
+
     # now do the actual upgrade
     for pkg in cache:
         if options.debug and pkg.is_upgradable:
             logging.debug("Checking: %s (%s)" % (
                 pkg.name, getattr(pkg.candidate, "origins", [])))
         if (pkg.is_upgradable and
-                not is_pkgname_in_blacklist(pkg.name, blacklisted_pkgs) and
-                is_allowed_origin(pkg.candidate, allowed_origins)):
-            try:
-                pkg.mark_upgrade()
-                if check_changes_for_sanity(cache, allowed_origins,
-                                            blacklisted_pkgs):
-                    # add to packages to upgrade
-                    pkgs_to_upgrade.append(pkg)
-                    # re-eval pkgs_kept_back as the resolver may fail to
-                    # directly upgrade a pkg, but that may work during
-                    # a subsequent operation, see debian bug #639840
-                    for pkgname in pkgs_kept_back:
-                        if (cache[pkgname].marked_install or
-                                cache[pkgname].marked_upgrade):
-                            pkgs_kept_back.remove(pkgname)
-                            pkgs_to_upgrade.append(cache[pkgname])
-                else:
-                    logging.debug("sanity check failed")
-                    rewind_cache(cache, pkgs_to_upgrade)
-                    pkgs_kept_back.append(pkg.name)
-            except SystemError as e:
-                # can't upgrade
-                logging.warning(
-                    _("package '%s' upgradable but fails to "
-                      "be marked for upgrade (%s)") % (pkg.name, e))
-                rewind_cache(cache, pkgs_to_upgrade)
-                pkgs_kept_back.append(pkg.name)
+              not is_pkgname_in_blacklist(pkg.name, blacklisted_pkgs) and
+              is_pkgname_in_whitelist(pkg.name, whitelisted_pkgs) and
+              is_allowed_origin(pkg.candidate, allowed_origins)):
+            try_to_upgrade(pkg,
+                           pkgs_to_upgrade,
+                           pkgs_kept_back,
+                           cache,
+                           allowed_origins,
+                           blacklisted_pkgs,
+                           whitelisted_pkgs)
     return pkgs_to_upgrade, pkgs_kept_back
 
 
+def get_dpkg_log_content(logfile_dpkg, install_start_time):
+    logging.debug("Extracting content from '%s' since '%s'" % (
+        logfile_dpkg, install_start_time))
+    content = []
+    found_start = False
+    with open(logfile_dpkg) as fp:
+        # read until we find the last "Log started: "
+        for line in fp.readlines():
+            # scan for the first entry we need (minimal-step mode
+            # creates a new stanza for each individual install)
+            if not found_start and line.startswith("Log started: "):
+                stanza_start = LoggingDateTime.from_string(
+                    line[len("Log started: "):-1])
+                if stanza_start >= install_start_time:
+                    found_start = True
+            if found_start:
+                content.append(line)
+    return "".join(content)
+
+
 def main(options, rootdir=""):
 
     # useful for testing
@@ -935,10 +1039,16 @@ def main(options, rootdir=""):
     # format (origin, archive), e.g. ("Ubuntu","dapper-security")
     allowed_origins = get_allowed_origins()
 
-    # pkgs that are (for some reason) not save to install
+    # pkgs that are (for some reason) not safe to install
     blacklisted_pkgs = get_blacklisted_pkgs()
     logging.info(_("Initial blacklisted packages: %s"),
                  " ".join(blacklisted_pkgs))
+
+    # install only these packages regardless of other upgrades available
+    whitelisted_pkgs = get_whitelisted_pkgs()
+    logging.info(_("Initial whitelisted packages: %s"),
+                 " ".join(whitelisted_pkgs))
+
     logging.info(_("Starting unattended upgrades script"))
 
     # display available origin
@@ -962,7 +1072,9 @@ def main(options, rootdir=""):
             try:
                 os.close(lockfd)
                 output = subprocess.check_output(
-                    ["dpkg", "--force-confold", "--configure", "-a"], env=env)
+                    ["dpkg", "--force-confold", "--configure", "-a"],
+                    env=env,
+                    universal_newlines=True)
             except subprocess.CalledProcessError as e:
                 output = e.output
             logging.warning(_("dpkg --configure -a output:\n%s" % output))
@@ -1000,7 +1112,7 @@ def main(options, rootdir=""):
                                if pkg.is_auto_removable])
     # find out about the packages that are upgradable (in a allowed_origin)
     pkgs_to_upgrade, pkgs_kept_back = calculate_upgradable_pkgs(
-        cache, options, allowed_origins, blacklisted_pkgs)
+        cache, options, allowed_origins, blacklisted_pkgs, whitelisted_pkgs)
     pkgs_to_upgrade.sort(key=lambda p: p.name)
     pkgs = "\n".join([pkg.name for pkg in pkgs_to_upgrade])
     logging.debug("pkgs that look like they should be upgraded: %s" % pkgs)
@@ -1069,16 +1181,19 @@ def main(options, rootdir=""):
         # redo the selection about the packages to upgrade based on the new
         # blacklist
         logging.debug("blacklist: %s" % blacklisted_pkgs)
+        # whitelist
+        logging.debug("whitelist: %s" % whitelisted_pkgs)
         # find out about the packages that are upgradable (in a allowed_origin)
-        if len(blacklisted_pkgs) > 0:
+        if len(blacklisted_pkgs) > 0 or len(whitelisted_pkgs) > 0:
             cache.clear()
             old_pkgs_to_upgrade = pkgs_to_upgrade[:]
             pkgs_to_upgrade = []
             for pkg in old_pkgs_to_upgrade:
-                logging.debug("Checking (blacklist): %s" % (pkg.name))
+                logging.debug("Checking the black and whitelist: %s" %
+                              (pkg.name))
                 pkg.mark_upgrade()
                 if check_changes_for_sanity(cache, allowed_origins,
-                                            blacklisted_pkgs):
+                                            blacklisted_pkgs, whitelisted_pkgs):
                     pkgs_to_upgrade.append(pkg)
                 else:
                     if not (pkg.name in pkgs_kept_back):
@@ -1132,25 +1247,25 @@ def main(options, rootdir=""):
     logging.info(_("Packages that will be upgraded: %s" % pkgs))
 
     # get log
-    now = datetime.datetime.now()
-    logfile_dpkg = os.path.join(
-        _get_logdir(), 'unattended-upgrades-dpkg_%s.log' % now.isoformat('_'))
+    logfile_dpkg = os.path.join(_get_logdir(), 'unattended-upgrades-dpkg.log')
 
     # only perform install step if we actually have packages to install
     pkg_install_success = None
     shutdown_lock = -1
+    install_start_time = datetime.datetime.now()
     if len(pkgs_to_upgrade) > 0:
         # lock for the shutdown check - its fine if the system
         # is shutdown while downloading but not so much while installing
         shutdown_lock = apt_pkg.get_lock("/var/run/unattended-upgrades.lock")
         # do install
         pkg_install_success = do_install(
-            cache, pkgs_to_upgrade, blacklisted_pkgs, options, logfile_dpkg)
+            cache, pkgs_to_upgrade, blacklisted_pkgs, whitelisted_pkgs, options, logfile_dpkg)
 
     # send a mail (if needed)
     if not options.dry_run:
+        log_content = get_dpkg_log_content(logfile_dpkg, install_start_time)
         send_summary_mail(
-            pkgs, pkg_install_success, pkgs_kept_back, mem_log, logfile_dpkg)
+            pkgs, pkg_install_success, pkgs_kept_back, mem_log, log_content)
 
     # FIXME: DRY violation, write_stamp_file() is used above as well
     # write timestamp file
@@ -1167,6 +1282,9 @@ if __name__ == "__main__":
     gettext.bindtextdomain(localesApp, localesDir)
     gettext.textdomain(localesApp)
 
+    # this ensures the commandline is logged in /var/log/apt/history.log
+    apt_pkg.config.set("Commandline::AsString", " ".join(sys.argv))
+
     # init the options
     parser = OptionParser()
     parser.add_option("-d", "--debug",
